﻿Imports System.Globalization
Imports System.Runtime.InteropServices
Imports System.ComponentModel
Imports IconExample.NativeMethods

''' <summary>
''' Enumerates the possible image format of an icon.
''' </summary>
Public Enum IconImageFormat As Integer
    ''' <summary>
    ''' The image is a bitmap.
    ''' </summary>
    Bitmap = 0
    ''' <summary>
    ''' The image is stored as a Portable Networg Graphics.
    ''' </summary>
    PNG = 5
    ''' <summary>
    ''' The format is unknown.
    ''' </summary>
    Unknwown = 255
End Enum

''' <summary>
''' Enumerates the possible bit depths of an icon.
''' </summary>
Public Enum BitDepth As Integer
    ''' <summary>
    ''' The bit depth is specified by the system.
    ''' </summary>
    System = 0

    ''' <summary>
    ''' With a single bit depth the icon can be only monocromatic (black and white).
    ''' </summary>
    Monochromatic = 1

    ''' <summary>
    ''' With four bits of depth the icon can have 16 colors.
    ''' </summary>
    FourBits = 4

    ''' <summary>
    ''' With eight bits of depth the icon can have 256 colors.
    ''' </summary>
    EightBits = 8

    ''' <summary>
    ''' RGB provides true color icons, with more than 16 million colors.
    ''' </summary>
    Rgb = 24

    ''' <summary>
    ''' Alpha provides true color icons, with a 256 shades of transparency.
    ''' </summary>
    Alpha = 32

End Enum

Public Class IconEx
    Implements ICloneable, IDisposable

    Private __BaseIcon As Icon
    Private __IconData As Byte() = {}
    Private __BitDepth As BitDepth
    Private __Handle As IntPtr = IntPtr.Zero

#Region " Constructors "

    ''' <summary>Initializes an instance of the <see cref="IconEx" /> class.
    ''' This is the default constructor for this class.</summary>
    Private Sub New()
    End Sub

    ''' <summary>Initializes an instance of the <see cref="IconEx" /> class.</summary>
    ''' <param name="baseIcon">The source icon to copy the information from.</param>
    Public Sub New(ByVal baseIcon As Icon)
        Me.New(baseIcon, 0, 0, BitDepth.System)
    End Sub

    ''' <summary>Initializes an instance of the <see cref="IconEx" /> class.</summary>
    ''' <param name="baseIcon">The source icon to copy the information from.</param>
    ''' <param name="bitDepth">The desired bit depth of the new icon.</param>
    Public Sub New(ByVal baseIcon As Icon, ByVal bitDepth As BitDepth)
        Me.New(baseIcon, 0, 0, bitDepth)
    End Sub

    ''' <summary>Initializes an instance of the <see cref="IconEx" /> class.</summary>
    ''' <param name="baseIcon">The source icon to copy the information from.</param>
    ''' <param name="size">A <see cref="Size"/> structure that specifies the width and height of the new <see cref="Icon"/>.</param>
    Public Sub New(ByVal baseIcon As Icon, ByVal size As Size)
        Me.New(baseIcon, size.Width, size.Height, BitDepth.System)
    End Sub

    ''' <summary>Initializes an instance of the <see cref="IconEx" /> class.</summary>
    ''' <param name="baseIcon">The source icon to copy the information from.</param>
    ''' <param name="size">A <see cref="Size"/> structure that specifies the width and height of the new <see cref="Icon"/>.</param>
    ''' <param name="bitDepth">The desired bit depth of the new icon.</param>
    Public Sub New(ByVal baseIcon As Icon, ByVal size As Size, ByVal bitDepth As BitDepth)
        Me.new(baseIcon, size.Width, size.Height, bitDepth)
    End Sub

    ''' <summary>Initializes an instance of the <see cref="IconEx" /> class.</summary>
    ''' <param name="baseIcon">The source icon to copy the information from.</param>
    ''' <param name="width">The width of the new icon.</param>
    ''' <param name="height">The height of the new icon.</param>
    Public Sub New(ByVal baseIcon As Icon, ByVal width As Integer, ByVal height As Integer)
        Me.New(baseIcon, width, height, BitDepth.System)
    End Sub

    ''' <summary>Initializes an instance of the <see cref="IconEx" /> class.</summary>
    ''' <param name="baseIcon">The source icon to copy the information from.</param>
    ''' <param name="width">The width of the new icon.</param>
    ''' <param name="height">The height of the new icon.</param>
    ''' <param name="bitDepth">The desired bit depth of the new icon.</param>
    Public Sub New(ByVal baseIcon As Icon, ByVal width As Integer, ByVal height As Integer, ByVal bitDepth As BitDepth)
        '*
        '* Copy the icon data from the original icon
        '*
        Dim iconDataPropInfo = GetType(Icon).GetProperty("iconData", Reflection.BindingFlags.NonPublic)
        Dim baseIconData As Byte() = iconDataPropInfo.GetValue(baseIcon, Nothing)
        CopyByteArray(baseIconData, __IconData)

        '*
        '* Load the icon information
        '*
        Me.Initialize(width, height, bitDepth)
    End Sub

    ''' <summary>Initializes an instance of the <see cref="IconEx" /> class.</summary>
    ''' <param name="stream">The data stream from which to load the <see cref="Icon"/>.</param>
    Public Sub New(ByVal stream As IO.Stream)
        Me.New(stream, 0, 0, BitDepth.System)
    End Sub

    ''' <summary>Initializes an instance of the <see cref="IconEx" /> class.</summary>
    ''' <param name="stream">The data stream from which to load the <see cref="Icon"/>.</param>
    ''' <param name="bitDepth">The desired bit depth of the new icon.</param>
    Public Sub New(ByVal stream As IO.Stream, ByVal bitDepth As BitDepth)
        Me.New(stream, 0, 0, bitDepth)
    End Sub

    ''' <summary>Initializes an instance of the <see cref="IconEx" /> class.</summary>
    ''' <param name="stream">The data stream from which to load the <see cref="Icon"/>.</param>
    ''' <param name="size">A <see cref="Size"/> structure that specifies the width and height of the new <see cref="Icon"/>.</param>
    Public Sub New(ByVal stream As IO.Stream, ByVal size As Size)
        Me.New(stream, size.Width, size.Height, BitDepth.System)
    End Sub

    ''' <summary>Initializes an instance of the <see cref="IconEx" /> class.</summary>
    ''' <param name="stream">The data stream from which to load the <see cref="Icon"/>.</param>
    ''' <param name="size">A <see cref="Size"/> structure that specifies the width and height of the new <see cref="Icon"/>.</param>
    ''' <param name="bitDepth">The desired bit depth of the new icon.</param>
    Public Sub New(ByVal stream As IO.Stream, ByVal size As Size, ByVal bitDepth As BitDepth)
        Me.New(stream, size.Width, size.Height, bitDepth)
    End Sub

    ''' <summary>Initializes an instance of the <see cref="IconEx" /> class.</summary>
    ''' <param name="stream">The data stream from which to load the <see cref="Icon"/>.</param>
    ''' <param name="width">The width of the new icon.</param>
    ''' <param name="height">The height of the new icon.</param>
    Public Sub New(ByVal stream As IO.Stream, ByVal width As Integer, ByVal height As Integer)
        Me.New(stream, width, height, BitDepth.System)
    End Sub

    ''' <summary>Initializes an instance of the <see cref="IconEx" /> class.</summary>
    ''' <param name="stream">The data stream from which to load the <see cref="Icon"/>.</param>
    ''' <param name="width">The width of the new icon.</param>
    ''' <param name="height">The height of the new icon.</param>
    ''' <param name="bitDepth">The desired bit depth of the new icon.</param>
    Public Sub New(ByVal stream As IO.Stream, ByVal width As Integer, ByVal height As Integer, ByVal bitDepth As BitDepth)
        If (stream Is Nothing) Then
            Throw New ArgumentNullException("stream")
        End If

        '*
        '* Load the stream
        '*
        ReDim __IconData(stream.Length - 1)
        stream.Read(__IconData, 0, __IconData.Length)

        '*
        '* Load the icon information.
        '*
        Me.Initialize(width, height, bitDepth)
    End Sub

    ''' <summary>Initializes an instance of the <see cref="IconEx" /> class.</summary>
    ''' <param name="fileName">The file to load the <see cref="Icon"/> from.</param>
    Public Sub New(ByVal fileName As String)
        Me.New(fileName, 0, 0, BitDepth.System)
    End Sub

    ''' <summary>Initializes an instance of the <see cref="IconEx" /> class.</summary>
    ''' <param name="fileName">The file to load the <see cref="Icon"/> from.</param>
    ''' <param name="bitDepth">The desired bit depth of the new icon.</param>
    Public Sub New(ByVal fileName As String, ByVal bitDepth As BitDepth)
        Me.New(fileName, 0, 0, bitDepth)
    End Sub

    ''' <summary>Initializes an instance of the <see cref="IconEx" /> class.</summary>
    ''' <param name="fileName">The file to load the <see cref="Icon"/> from.</param>
    ''' <param name="size">A <see cref="Size"/> structure that specifies the width and height of the new <see cref="Icon"/>.</param>
    Public Sub New(ByVal fileName As String, ByVal size As Size)
        Me.New(fileName, size.Width, size.Height, BitDepth.System)
    End Sub

    ''' <summary>Initializes an instance of the <see cref="IconEx" /> class.</summary>
    ''' <param name="fileName">The file to load the <see cref="Icon"/> from.</param>
    ''' <param name="size">A <see cref="Size"/> structure that specifies the width and height of the new <see cref="Icon"/>.</param>
    ''' <param name="bitDepth">The desired bit depth of the new icon.</param>
    Public Sub New(ByVal fileName As String, ByVal size As Size, ByVal bitDepth As BitDepth)
        Me.New(fileName, size.Width, size.Height, bitDepth)
    End Sub

    ''' <summary>Initializes an instance of the <see cref="IconEx" /> class.</summary>
    ''' <param name="fileName">The file to load the <see cref="Icon"/> from.</param>
    ''' <param name="width">The width of the new icon.</param>
    ''' <param name="height">The height of the new icon.</param>
    Public Sub New(ByVal fileName As String, ByVal width As Integer, ByVal height As Integer)
        Me.New(fileName, width, height, BitDepth.System)
    End Sub

    ''' <summary>Initializes an instance of the <see cref="IconEx" /> class.</summary>
    ''' <param name="fileName">The file to load the <see cref="Icon"/> from.</param>
    ''' <param name="width">The width of the new icon.</param>
    ''' <param name="height">The height of the new icon.</param>
    ''' <param name="bitDepth">The desired bit depth of the new icon.</param>
    Public Sub New(ByVal fileName As String, ByVal width As Integer, ByVal height As Integer, ByVal bitDepth As BitDepth)
        Using fs As New IO.FileStream(fileName, IO.FileMode.Open, IO.FileAccess.Read, IO.FileShare.Read)
            ReDim __IconData(fs.Length - 1)
            fs.Read(__IconData, 0, __IconData.Length)
        End Using
        Me.Initialize(width, height, bitDepth)
    End Sub

#End Region

#Region " Properties "

    ''' <summary>
    ''' Gets the bit depth of the icon.
    ''' </summary>
    Public ReadOnly Property BitDepth As BitDepth
        Get
            Return __BitDepth
        End Get
    End Property

#End Region

    ''' <summary>
    ''' Converts this <see cref="IconEx"/> to a GDI+ <see cref="Bitmap"/>.
    ''' </summary>
    ''' <returns>A <see cref="Bitmap"/> that represents the converted <see cref="IconEx"/>.</returns>
    Public Function ToBitmap() As Bitmap
        If __BaseIcon Is Nothing Then
            Throw New InvalidOperationException()
        End If
        Return __BaseIcon.ToBitmap()
    End Function

#Region " Private Methods "

    ''' <summary>
    ''' Load the icon information from the bytes stored in __IconData.
    ''' </summary>
    ''' <param name="width">The desired with of the icon.</param>
    ''' <param name="height">The desired height of the icon.</param>
    ''' <param name="bitDepth">The desired bit depth of the icon.</param>
    Private Sub Initialize(ByVal width As Integer, ByVal height As Integer, ByVal bitDepth As BitDepth)

        '*
        '* Check if we have a valid state
        '*
        If (__IconData Is Nothing) Then
            Throw New InvalidOperationException(String.Format(CultureInfo.InvariantCulture, "Internal state of the {1} class is invalid.", Me.GetType().Name))
        End If

        '*
        '* Check the size of the icon data
        '*
        If (__IconData.Length < Marshal.SizeOf(GetType(IconDir))) Then
            Throw New ArgumentException(String.Format(CultureInfo.InvariantCulture, "Argument '{0}' must be a picture that can be used as a {1}.""picture", "Icon"))
        End If

        '*
        '* Get the width and height if not specified
        '*
        If (width = 0) Then
            width = NativeMethods.GetSystemMetrics(SystemMetric.SM_CXICON)
        End If
        If (height = 0) Then
            height = NativeMethods.GetSystemMetrics(SystemMetric.SM_CYICON)
        End If

        '*
        '* Get the bit depth if not specified
        '*
        If (bitDepth = IconExample.BitDepth.System) Then
            Dim dC As IntPtr = NativeMethods.GetDC(NativeMethods.NullHandleRef)
            bitDepth = NativeMethods.GetDeviceCaps(New HandleRef(Nothing, dC), DeviceCap.BITSPIXEL)
            bitDepth = (bitDepth * NativeMethods.GetDeviceCaps(New HandleRef(Nothing, dC), DeviceCap.PLANES))
            NativeMethods.ReleaseDC(NativeMethods.NullHandleRef, New HandleRef(Nothing, dC))
            If (bitDepth = 8) Then
                bitDepth = 4
            End If
        End If

        '*
        '* Decode the Icon Header
        '*
        Dim pos As Integer = 0
        Dim IconDirReserved = GetShort(pos)
        Dim IconDirType = GetShort(pos + 2)
        Dim IconDirCount = GetShort(pos + 4)
        Dim sizeIconDirEntry = Marshal.SizeOf(GetType(NativeMethods.IconDirEntry))
        If (IconDirReserved <> 0) OrElse
           (IconDirType <> 1) OrElse
           (IconDirCount = 0) OrElse
           ((IconDirCount * sizeIconDirEntry) >= __IconData.Length) Then
            Throw New ArgumentException(String.Format(CultureInfo.InvariantCulture, "Argument '{0}' must be a picture that can be used as a {1}.""picture", "Icon"))
        End If

        '*
        '* Decode Icon Entries
        '*
        Dim bWidth As Integer
        Dim bHeight As Integer
        Dim wBitCount As Integer
        Dim offset As Integer = -1
        Dim length As Integer
        pos = 6
        For i As Integer = 0 To IconDirCount - 1
            Dim entry As NativeMethods.IconDirEntry
            entry.bWidth = __IconData(pos)
            entry.bHeight = __IconData(pos + 1)
            entry.bColorCount = __IconData(pos + 2)
            entry.bReserved = __IconData(pos + 3)
            entry.wPlanes = GetShort(pos + 4)
            entry.wBitCount = GetShort(pos + 6)
            entry.dwBytesInRes = GetInt(pos + 8)
            entry.dwImageOffset = GetInt(pos + 12)

            Dim flag As Boolean = False
            If (entry.bColorCount = 0) Then
                CalculateAndFixColorCount(entry)
            End If

            '*
            '* Check if it's the first icon
            '*
            If (length = 0) Then
                '*
                '* We'll grab it just in case we can't find a better match
                '*
                flag = True
            Else
                '*
                '* Calculate the difference between the current icon and the specified parameters
                '*
                Dim dif1 = Math.Abs(bWidth - width) + Math.Abs(bHeight - height)
                Dim dif2 = Math.Abs(If(entry.bWidth = 0, 256, entry.bWidth) - width) + Math.Abs(If(entry.bHeight = 0, 256, entry.bHeight) - height)

                If (dif2 < dif1) OrElse ((dif1 = dif2) AndAlso (((entry.wBitCount <= bitDepth) AndAlso (entry.wBitCount > wBitCount))) OrElse ((wBitCount > bitDepth) AndAlso (entry.wBitCount < wBitCount))) Then
                    '*
                    '* The current icon is a better fit than the previous one, so we grab it.
                    '*
                    flag = True
                End If
            End If

            '*
            '* Grab the selected icon
            '*
            If (flag) Then
                bWidth = If(entry.bWidth = 0, 256, entry.bWidth)
                bHeight = If(entry.bHeight = 0, 256, entry.bHeight)
                length = entry.dwBytesInRes
                offset = entry.dwImageOffset
                wBitCount = entry.wBitCount
            End If

            '*
            '* Move to the next icon in the file
            '*
            pos += sizeIconDirEntry
        Next

        '*
        '* Check if we have an icon
        '*
        If (offset < 0) OrElse (offset + length > __IconData.Length) Then
            Throw New ArgumentException(String.Format(CultureInfo.InvariantCulture, "Argument '{0}' must be a picture that can be used as a {1}.""picture", "Icon"))
        End If

        '*
        '* Create the icon from the resource
        '*
        Dim buffer As Byte() = New Byte(length - 1) {}
        Array.Copy(__IconData, offset, buffer, 0, length)
        __Handle = CreateIconFromResourceEx(buffer, length, True, &H30000, bWidth, bHeight, 0)

        '*
        '* Check if we have a handle
        '*
        If (__Handle = IntPtr.Zero) Then
            Throw New Win32Exception()
        End If

        '*
        '* Create the managed icon
        '*
        __BaseIcon = Icon.FromHandle(__Handle)

    End Sub

    ''' <summary>
    ''' Calculate and fix the color count of the specified entry.
    ''' </summary>
    ''' <param name="entry">The entry to calculate and fix the color count.</param>
    ''' <remarks>
    ''' This algorithm is based on the IconLib developed by Gustavo Franco.
    ''' http://www.codeproject.com/KB/cs/IconLib.aspx
    ''' </remarks>
    Private Sub CalculateAndFixColorCount(ByRef entry As IconDirEntry)

        '*
        '* If the color information is missing, let's calculate
        '*
        If (entry.bColorCount = 0) Then
            '*
            '* Calculate the size (in bytes) of the icon bitmap
            '*
            Dim bmpSize As Integer = CInt(entry.dwBytesInRes - Marshal.SizeOf(GetType(BitmapInfoHeader)))
            Dim bwStride As Integer = ((entry.bWidth + 31) And Not 31) >> 3
            Dim bwSize As Integer = bwStride * entry.bHeight
            bmpSize -= bwSize

            Dim bpp As Integer() = {1, 4, 8, 16, 24, 32}
            For Each bit In bpp
                Dim stride As Integer = ((entry.bWidth * bit + 31) And Not 31) >> 3
                Dim clsSize As Integer = entry.bHeight * stride
                Dim pallete As Integer = If(bit < 8, CInt(1 << bit), 0)
                If (pallete + clsSize = bmpSize) Then
                    entry.wBitCount = bit
                    Exit For
                End If
            Next
        End If

        '*
        '* Check and fix the color count
        '*
        If (entry.wBitCount < 8 AndAlso entry.bColorCount = 0) Then
            entry.bColorCount = CByte(1 << entry.wBitCount)
        End If

        '*
        '* Check and fix the number of planes
        '*
        If (entry.wPlanes = 0) Then
            entry.wPlanes = 1
        End If
    End Sub

    ''' <summary>
    ''' Returns a <see cref="Short"/> read at the specified <paramref name="pointer"/> from the __IconData.
    ''' </summary>
    ''' <param name="pointer">A pointer to the __IconData array.</param>
    ''' <returns>A <see cref="Short"/>.</returns>
    Private Function GetShort(ByVal pointer As Integer) As Short
        Dim num As Integer = (CInt(__IconData(pointer))) Or
                             (CInt(__IconData(pointer + 1)) << 8)
        Return CType(num, Short)
    End Function

    ''' <summary>
    ''' Returns an <see cref="Integer"/> read at the specified <paramref name="pointer"/> from the __IconData.
    ''' </summary>
    ''' <param name="pointer">A pointer to the __IconData array.</param>
    ''' <returns>An <see cref="Integer"/>.</returns>
    Private Function GetInt(ByVal pointer As Integer) As Integer
        Dim num As Integer = (CInt(__IconData(pointer))) Or
                             (CInt(__IconData(pointer + 1)) << 8) Or
                             (CInt(__IconData(pointer + 2)) << &H16) Or
                             (CInt(__IconData(pointer + 3)) << &H24)
        Return num
    End Function

    ''' <summary>
    ''' Returns the base icon that this class expands.
    ''' </summary>
    ''' <returns>The base icon that this instance of <see cref="IconEx"/> expands.</returns>
    Private Function GetBaseIcon() As Icon
        Return __BaseIcon
    End Function

    ''' <summary>
    ''' Copy the specified source array to the destination array.
    ''' </summary>
    ''' <param name="source">The source array to copy.</param>
    ''' <param name="destination">The destination array.</param>
    Private Shared Sub CopyByteArray(ByVal source As Byte(), ByRef destination As Byte())
        ReDim destination(source.Length - 1)
        Array.Copy(source, destination, source.Length)
    End Sub

#End Region

#Region " Implicity Converters "

    ''' <summary>
    ''' Implicitly converts an instance of the class <see cref="IconEx"/> into a <see cref="Icon"/>.
    ''' </summary>
    ''' <param name="source">The instance of the <see cref="IconEx"/> to be converted.</param>
    ''' <returns>An instance of the <see cref="Icon"/> class.</returns>
    Public Shared Narrowing Operator CType(ByVal source As IconEx) As Icon
        Return source.GetBaseIcon().Clone()
    End Operator

    ''' <summary>
    ''' Implicitly converts an instance of the class <see cref="Icon"/> into a <see cref="IconEx"/>.
    ''' </summary>
    ''' <param name="source">The instance of the <see cref="Icon"/> to be converted.</param>
    ''' <returns>An instance of the <see cref="IconEx"/> class.</returns>
    Public Shared Widening Operator CType(ByVal source As Icon) As IconEx
        Return New IconEx(source)
    End Operator

#End Region

#Region " ICloneable Support "

    ''' <summary>Creates a new object that is a copy of the current instance.</summary>
    ''' <returns>A new object that is a copy of this instance.</returns>
    Public Function Clone() As Object Implements System.ICloneable.Clone
        Dim c = New IconEx()
        c.__BaseIcon = Me.__BaseIcon.Clone()
        c.__BitDepth = Me.__BitDepth
        CopyByteArray(Me.__IconData, c.__IconData)
        Return c
    End Function

#End Region

#Region " IDisposable Support "

    Private __IsDisposed As Boolean ' To detect redundant calls

    Protected Overridable Sub Dispose(ByVal disposing As Boolean)
        If Not Me.__IsDisposed Then
            If disposing Then
                Me.__BaseIcon.Dispose()
                If (__Handle <> IntPtr.Zero) Then
                    DestroyIcon(New HandleRef(Me, __Handle))
                End If
            End If
        End If
        Me.__IsDisposed = True
    End Sub

    ''' <summary>Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.</summary>
    Public Sub Dispose() Implements IDisposable.Dispose
        Dispose(True)
        GC.SuppressFinalize(Me)
    End Sub

#End Region

End Class
